#ifndef MYX_CORE_ENUM_BITWISE_OPERATIONS_HPP_
#define MYX_CORE_ENUM_BITWISE_OPERATIONS_HPP_

#pragma once

#include <type_traits>

namespace myx {

namespace core {

template< typename Enum >
struct EnableBitMaskOperators
{
	static const bool kEnable = false;
};

template< typename Enum >
typename std::enable_if< EnableBitMaskOperators< Enum >::kEnable, Enum >::type
operator ~( Enum e )
{
	using Underlying = typename std::underlying_type< Enum >::type;
	return( static_cast< Enum >( ~static_cast< Underlying >( e ) ) );
}


template< typename Enum >
typename std::enable_if< EnableBitMaskOperators< Enum >::kEnable, Enum >::type
operator |( Enum lhs, Enum rhs )
{
	using Underlying = typename std::underlying_type< Enum >::type;
	return( static_cast< Enum >( static_cast< Underlying >( lhs ) | static_cast< Underlying >( rhs ) ) );
}


template< typename Enum >
typename std::enable_if< EnableBitMaskOperators< Enum >::kEnable, typename std::add_lvalue_reference< Enum >::type >::type
operator |=( Enum& lhs, Enum rhs )
{
	using Underlying = typename std::underlying_type< Enum >::type;
	lhs              = static_cast< Enum >( static_cast< Underlying >( lhs ) | static_cast< Underlying >( rhs ) );
	return( lhs );
}


template< typename Enum >
typename std::enable_if< EnableBitMaskOperators< Enum >::kEnable, Enum >::type
operator &( Enum lhs, Enum rhs )
{
	using Underlying = typename std::underlying_type< Enum >::type;
	return( static_cast< Enum >( static_cast< Underlying >( lhs ) & static_cast< Underlying >( rhs ) ) );
}


template< typename Enum >
typename std::enable_if< EnableBitMaskOperators< Enum >::kEnable, typename std::add_lvalue_reference< Enum >::type >::type
operator &=( Enum& lhs, Enum rhs )
{
	using Underlying = typename std::underlying_type< Enum >::type;
	lhs              = static_cast< Enum >( static_cast< Underlying >( lhs ) & static_cast< Underlying >( rhs ) );
	return( lhs );
}

} // namespace core

} // namespace myx

/**
 * @brief Макрос, предоставляющий возможность выполнять битовые операции в enum class
 *
 *  Источник: http://blog.bitwigglers.org/using-enum-classes-as-type-safe-bitmasks/
 *  Пример использования:
 *
 *  namespace ns {
 *  enum class Permissions
 *  {
 *     Readable   = 0x4,
 *     Writeable  = 0x2,
 *     Executable = 0x1
 *  };
 *  }
 *  ENABLE_BITMASK_OPERATORS(ns::Permissions)
 */

// NOLINTNEXTLINE(cppcoreguidelines-macro-usage)
#define ENABLE_BITMASK_OPERATORS( x ) \
	template<> \
	struct myx::core::EnableBitMaskOperators< x > \
	{ \
		static const bool kEnable = true; \
	};

#endif // ifndef MYX_CORE_ENUM_BITWISE_OPERATIONS_HPP_
